# -*- coding:utf-8 -*-
"""
This module contains a set of functions to deal with geodatadecode
commonly encoded variables
Created on 15/10/2020
by gabrielmcf
license: GPL V3 or Later
"""

__docformat__ = "restructuredtext en"
import pandas as pd
from geobr import read_municipality


def add_data_to_municipality(
    counts,
    map_year=2019,
    codmun_col="CODMUNRES",
    title_cols=["SEXO", "IDADE_ANOS"],
    value_col="COUNTS",
    nan_string="nan",
):

    """
    Adiciona dados de mortalidade aos seus respectivos municípios. Gera um GeoDataFrame do GeoPandas.
    :param counts: dataframe contendo os dados a serem agregados.
    :param map_year: ano do mapa a ser usado (biblioteca geobr).
    :param codmun_col: coluna com geocode do município
    :param title_cols: colunas que serão utilizadas para formar o título das colunas no GeoDataFrame.
    :param value_col: coluna com o valor a ser adicionado ao GeoDataFrame
    :return:
    """

    # Extrai código do estado dos municípios.
    # 2 primeiros dígitos do código são o estado
    states = (
        counts[counts[codmun_col] != nan_string][codmun_col]
        .apply(lambda x: str(x)[:2])
        .unique()
    )
    geo_df = read_municipality(code_muni=states[0], year=map_year)

    all_fields = [codmun_col] + title_cols + [value_col]

    if len(states) > 1:
        for state in states[1:]:
            geo_df = geo_df.append(read_municipality(code_muni=state, year=map_year))

    column_names = column_name_list(counts, title_cols)
    geo_df[column_names] = 0

    for i, mun in geo_df.iterrows():
        data = counts[
            (counts[codmun_col] == mun["code_muni"]) & (counts[value_col] > 0.0)
        ]
        for _, item in data.iterrows():
            geo_df.loc[i, column_name(item, title_cols)] = item[value_col]

    return geo_df.fillna(0)


def column_name_list(df, title_cols):
    unique_indices = pd.DataFrame(df.groupby(title_cols).indices.keys())
    titles = unique_indices.apply(column_name, axis=1)
    return titles.tolist()


def column_name(item, title_cols=None):
    if title_cols:
        name = item[title_cols[0]]
        if len(title_cols) > 1:
            for col in title_cols[1:]:
                name = f"{name}-{item[col]}"
    else:
        name = item[0]
        if len(item) > 1:
            for col in item[1:]:
                name = f"{name}-{col}"

    return name



__author__ = "fccoelho"

import unittest

import pandas as pd

from pysus.online_data.SIA import download

unittest.skip("too slow to run om travis")


class SIATestCase(unittest.TestCase):
    def test_download_after_2008(self):
        data = download("to", 2015, 12)
        # print(data)
        self.assertGreater(len(data), 0)
        for df in data:
            if df is None:
                continue
            self.assertIn("PA_CODUNI", df.columns)
            self.assertIn("CODUNI", df.columns)
            self.assertIsInstance(df, pd.DataFrame)
            self.assertIsInstance(df, pd.DataFrame)

    def test_download_before_2008(self):
        data = download("mg", 2005, 8)
        self.assertWarns(UserWarning)
        for df in data:
            if df is None:
                continue
            self.assertGreater(len(df), 0)
            self.assertIn("PA_CODUNI", df.columns)
            self.assertIsInstance(df, pd.DataFrame)

    @unittest.expectedFailure
    def test_download_before_1994(self):
        df1, df2 = download("RS", 1993, 12)

    def test_download_one(self):
        data = download("se", 2020, 10, group="PS")

        for df in data:
            if df is None:
                continue
            self.assertGreater(len(df), 0)
            self.assertIn("CNS_PAC", df.columns)
            self.assertIsInstance(df, pd.DataFrame)

    def test_download_many(self):
        dfs = download("PI", 2018, 3, group=["aq", "AM", "atd"])
        self.assertEqual(len(dfs), 3)
        df1, df2, df3 = dfs
        self.assertIsNone(df1)
        if df1 is None:
            return
        self.assertIsInstance(df1, pd.DataFrame)
        self.assertIsInstance(df2, pd.DataFrame)
        self.assertIsInstance(df2, pd.DataFrame)
        self.assertGreater(len(df1), 0)
        self.assertGreater(len(df2), 0)
        self.assertGreater(len(df3), 0)
        self.assertIn("AP_CODUNI", df1.columns)
        self.assertIn("AP_CODUNI", df2.columns)
        self.assertIn("AP_CODUNI", df3.columns)
        self.assertIn("AQ_CID10", df1.columns)
        self.assertIn("AM_PESO", df2.columns)
        self.assertIn("ATD_CARACT", df3.columns)

    def test_download_missing(self):
        dfs = download("MS", 2006, 5, group=["PA", "SAD"])
        assert len(dfs) == 2
        self.assertIsNone(dfs[0])
        self.assertIsNone(dfs[1])


if __name__ == "__main__":
    unittest.main()



"""
Download SINASC data from DATASUS FTP server
Created on 01/11/17
by fccoelho
license: GPL V3 or Later
"""
import os
import warnings
import pandas as pd

from ftplib import FTP
from loguru import logger

from pysus.online_data import CACHEPATH
from pysus.utilities.readdbc import read_dbc

warnings.filterwarnings("ignore", message=".*initial implementation of Parquet.*")


def download(state, year, cache=True):
    """
    Downloads data directly from Datasus ftp server
    :param state: two-letter state identifier: MG == Minas Gerais
    :param year: 4 digit integer
    :return: pandas dataframe
    """
    assert len(str(year)) == 4
    state = state.upper()

    if year < 1994:
        raise ValueError("SINASC does not contain data before 1994")

    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")

    if year >= 1996:
        ftp.cwd("/dissemin/publicos/SINASC/NOV/DNRES")
        logger.debug("Changing FTP work dir to: /dissemin/publicos/SINASC/NOV/DNRES")
        fname = "DN{}{}.DBC".format(state, year)

    else:
        ftp.cwd("/dissemin/publicos/SINASC/ANT/DNRES")
        logger.debug("Changing FTP work dir to: /dissemin/publicos/SINASC/ANT/DNRES")
        fname = "DNR{}{}.DBC".format(state, str(year)[-2:])

    cachefile = os.path.join(CACHEPATH, "SINASC_" + fname.split(".")[0] + "_.parquet")

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)
    df = read_dbc(fname, encoding="iso-8859-1")

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def get_available_years(state):
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")

    ftp.cwd("/dissemin/publicos/SINASC/ANT/DNRES")
    logger.debug("Changing FTP work dir to: /dissemin/publicos/SINASC/ANT/DNRES")
    res = ftp.nlst(f"DNR{state}*.*")

    ftp.cwd("/dissemin/publicos/SINASC/NOV/DNRES")
    logger.debug("Changing FTP work dir to: /dissemin/publicos/SINASC/NOV/DNRES")
    res += ftp.nlst(f"DN{state}*.*")
    
    return res



import unittest
from ftplib import FTP
import pandas as pd
from pysus.online_data.SIA import download, check_file_split

class SIATestCase(unittest.TestCase):
    def test_check_split_filenames(self):
        ftp = FTP("ftp.datasus.gov.br")
        ftp.login()
        ftp.cwd("/dissemin/publicos/SIASUS/200801_/Dados")
        names = check_file_split('PASP2012.dbc', ftp)
        assert len(names) == 3
        assert 'PASP2012b.dbc' in names

    @unittest.skip  # Takes a long time to complete
    def test_download_large_PA(self):
        res = download('SP', 2020, 12, group=['PA'])
        if isinstance(res, pd.DataFrame):
            assert not res.empty
        else:
            pass


if __name__ == '__main__':
    unittest.main()



import datetime
import os
from pathlib import Path
import shutil
import unittest
from glob import glob
from pathlib import Path

import numpy as np
import pandas as pd

from pysus.online_data.SINAN import download, list_diseases, download_all_years_in_chunks
from pysus.preprocessing.sinan import read_sinan_dbf

PATH_ROOT = Path(__file__).resolve().parent


class TestSINANDownload(unittest.TestCase):
    def test_download(self):
        df = download(year=2007, disease="Botulismo")
        self.assertIsInstance(df, pd.DataFrame)

    def test_filename_only(self):
        fname = download(year=2015, disease="Botulismo", return_chunks=True)
        self.assertIsInstance(fname, str)
        self.assertTrue(os.path.exists(fname))
        shutil.rmtree(fname, ignore_errors=True)

    def test_fetch_viol_dom(self):
        df = download(year=2011, disease="Hantavirose")
        self.assertIsInstance(df, pd.DataFrame)

    def test_fetch_cancer_prelim(self):
        df = download(year=2022, disease="Cancer")
        self.assertIsInstance(df, pd.DataFrame)

    def test_fetch_sifilis(self):
        self.assertRaises(
            Exception, download(year=2021, disease="Sífilis Adquirida")
        )

    def test_fetch_sifilis_gestante(self):
        df = download(year=2021, disease="Sífilis em Gestante")
        self.assertIsInstance(df, pd.DataFrame)

    def test_lista_agravos(self):
        lista = list_diseases()
        self.assertIsInstance(lista, list)
        self.assertGreater(len(lista), 0)

    def test_chunked_df_size(self):
        df1 = download(year=2018, disease='Chikungunya')
        s1 = len(df1)
        del df1
        fn = download(year=2018, disease='Chikungunya', return_chunks=True)
        for i, f in enumerate(glob(f'{fn}/*.parquet')):
            if i == 0:
                df2 = pd.read_parquet(f)
            else:
                df2 = pd.concat([df2, pd.read_parquet(f)], ignore_index=True)
        self.assertEqual(s1, df2.shape[0])
        shutil.rmtree(fn, ignore_errors=True)

    def test_download_all_dbfs_for_zika(self):
        download_all_years_in_chunks('zika')
        self.assertTrue(Path('/tmp/pysus/ZIKABR16.parquet').exists())
        self.assertTrue(Path('/tmp/pysus/ZIKABR17.parquet').exists())
        self.assertTrue(Path('/tmp/pysus/ZIKABR18.parquet').exists())
        self.assertTrue(Path('/tmp/pysus/ZIKABR19.parquet').exists())
        self.assertTrue(Path('/tmp/pysus/ZIKABR20.parquet').exists())

class TestSinanDBF(unittest.TestCase):
    dbf_name = PATH_ROOT / "EPR-2016-06-01-2016.dbf"

    def test_read_dbf(self):
        df = read_sinan_dbf(self.dbf_name, encoding="latin-1")
        self.assertTrue(self.dbf_name.exists())
        self.assertIsInstance(df, pd.DataFrame)
        for cname in df.columns:
            if cname.startswith("DT_"):
                self.assertIsInstance(df[cname][0], datetime.date)
            elif cname.startswith("SEM"):
                self.assertLessEqual(df[cname][0], 52)
                self.assertIsInstance(df[cname][0], (int, np.int64))
            elif cname.startswith(("NU", "ID")):
                if cname == "ID_AGRAVO":
                    continue
                self.assertIsInstance(
                    df[cname][0],
                    (int, float, np.int64),
                    msg="Failed on column {}, type:{}".format(
                        cname, type(df[cname][0])
                    ),
                )

    def test_type_convertion(self):
        df = read_sinan_dbf(self.dbf_name, encoding="latin-1")
        self.assertTrue(self.dbf_name.exists())
        assert not all(df.dtypes == "object")


if __name__ == "__main__":
    unittest.main()



import shutil
from ftplib import FTP
from pathlib import Path
from loguru import logger

from pysus.online_data import (
    _fetch_file,
    chunk_dbfiles_into_parquets,
    parquets_to_dataframe,
)


agravos = {
    "Animais Peçonhentos": "ANIM",
    "Botulismo": "BOTU",
    "Cancer": "CANC",
    "Chagas": "CHAG",
    "Chikungunya": "CHIK",
    "Colera": "COLE",
    "Coqueluche": "COQU",
    "Contact Communicable Disease": "ACBI",
    "Acidentes de Trabalho": "ACGR",
    "Dengue": "DENG",
    "Difteria": "DIFT",
    "Esquistossomose": "ESQU",
    "Febre Amarela": "FAMA",
    "Febre Maculosa": "FMAC",
    "Febre Tifoide": "FTIF",
    "Hanseniase": "HANS",
    "Hantavirose": "HANT",
    "Hepatites Virais": "HEPA",
    "Intoxicação Exógena": "IEXO",
    "Leishmaniose Visceral": "LEIV",
    "Leptospirose": "LEPT",
    "Leishmaniose Tegumentar": "LTAN",
    "Malaria": "MALA",
    "Meningite": "MENI",
    "Peste": "PEST",
    "Poliomielite": "PFAN",
    "Raiva Humana": "RAIV",
    "Sífilis Adquirida": "SIFA",
    "Sífilis Congênita": "SIFC",
    "Sífilis em Gestante": "SIFG",
    "Tétano Acidental": "TETA",
    "Tétano Neonatal": "TETN",
    "Tuberculose": "TUBE",
    "Violência Domestica": "VIOL",
    "Zika": "ZIKA",
}


def list_diseases():
    """List available diseases on SINAN"""
    return list(agravos.keys())


def get_available_years(disease, return_path=False):
    """
    Fetch available years for data related to specific disease
    :param disease: Disease name. See `SINAN.list_diseases` for valid names
    :param return_path: If set to True, returns the entire Path of the datasets
                        in the FTP Server. Used to remove the discrimination of
                        FINAIS and PRELIM while downloading the datasets.
    :return: A list of DBC files from a specific disease found in the FTP Server.
    """
    logger.warning(
        "Now SINAN tables are no longer split by state. Returning countrywide years"
    ) #legacy

    fpath = "/dissemin/publicos/SINAN/DADOS/FINAIS"
    ppath = "/dissemin/publicos/SINAN/DADOS/PRELIM"
    disease = check_case(disease)

    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")

    dbcs = []

    ftp.cwd(fpath)
    logger.debug(f"Changing FTP work dir to: {fpath}")
    for dbc in ftp.nlst(f"{agravos[disease]}BR*.dbc"):
        if return_path:
            dbcs.append(f"{fpath}/{dbc}")
        else:
            dbcs.append(dbc)
        
    ftp.cwd(ppath)
    logger.debug(f"Changing FTP work dir to: {ppath}")
    for dbc in ftp.nlst(f"{agravos[disease]}BR*.dbc"):
        if return_path:
            dbcs.append(f"{ppath}/{dbc}")
        else:
            dbcs.append(dbc)

    return dbcs


def download(disease, year, return_chunks=False, data_path="/tmp/pysus"):
    """
    Downloads SINAN data directly from Datasus ftp server.
    :param disease: Disease according to `agravos`.
    :param year: 4 digit integer.
    :param return_chunks: If set to True, download the data in parquet chunks.
    :param data_path: The directory where the chunks will be downloaded to.
    @note The data will be downloaded either return_chunks is set True or False,
          the difference between the two is that setting to False will read the
          parquet chunks, return as a DataFrame and clean after read.
    :return: Default behavior returns a Pandas DataFrame.
    """
    disease = check_case(disease)
    year2 = str(year)[-2:].zfill(2)
    dis_code = agravos[disease]
    fname = f"{dis_code}BR{year2}.dbc"
    years = get_available_years(disease) #legacy
    
    #Returns a list with all the DBC files found with their path,
    # enabling the user to download all the DBCs available in both
    # FINAIS and PRELIM directories 
    fyears = get_available_years(disease, return_path=True)

    first_year = [f.split(".")[0][-2:] for f in years][
        0
    ]

    if not years or fname not in years:
        raise Exception(f"No data found for this request. Available data for {disease}: \n{years}")

    if year2 < first_year: #legacy
        raise ValueError(f"SINAN does not contain data before {first_year}")

    logger.warning(
        "Now SINAN tables are no longer split by state. Returning country table" 
    ) #legacy
    #Generate the path to be downloaded from the FTP Server
    pname = next(p for p in fyears if fname in p)
    sus_path = "/".join(pname.split("/")[:-1])

    #Create the path where the data will be downloaded locally
    data_path = Path(data_path)
    data_path.mkdir(exist_ok=True, parents=True)
    logger.debug(f"{data_path} directory created.")

    out = Path(data_path) / fname
    dbf = Path(f"{str(out)[:-4]}.dbf")

    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")

    if not Path(out).exists():
        logger.debug(f"{fname} file not found. Proceeding to download..")
        try:
            _fetch_file(fname, sus_path, "DBC", return_df=False, data_path=data_path)
            logger.info(f"{fname} downloaded at {data_path}")

        except Exception as e:
            logger.error(e)

    try:
        partquet_dir = chunk_dbfiles_into_parquets(str(out))

        if not return_chunks:
            df = parquets_to_dataframe(partquet_dir, clean_after_read=True)
            return df

        return partquet_dir

    except Exception as e:
        logger.error(e)

    finally:
        out.unlink(missing_ok=True)
        dbf.unlink(missing_ok=True)
        Path(fname).unlink(missing_ok=True)
        Path(f'{fname[:-4]}.dbf').unlink(missing_ok=True)
        logger.debug("🧹 Cleaning data residues")


def download_all_years_in_chunks(disease, data_dir="/tmp/pysus"):
    """
    Download all DBFs found in datasus, given a disease, in chunks.
    An output path can be defined.
    `pysus.online_data.parquets_to_dataframe()` can read these parquets.
    :param disease: A disease according to `agravos`.
    :param data_dir: Output parquet path.
    """
    disease = check_case(disease)
    parquets = []

    available_years = get_available_years(disease, return_path=True)

    if available_years:
        for dbc in available_years:
            year = dbc.split('.dbc')[0][-2:]

            parquet_dir = download(
                disease = disease, 
                year = year, 
                return_chunks = True,
                data_path = data_dir
            )

            parquets.append(parquet_dir)

    return parquets


def check_case(disease):
    try:
        assert disease in agravos
    except AssertionError:
        try:
            assert disease.title()
            disease = disease.title()
        except AssertionError:
            print(
                f"Disease {disease.title()} is not available in SINAN.\n"
                "Available diseases: {list_diseases()}"
            )
    return disease



import unittest

from pysus.online_data.SIH import download


@unittest.skip("Waiting for Rio de Janeiro data on database demo.")
class SIHTestCase(unittest.TestCase):
    def test_download_pre_2008(self):
        df = download("AC", 2006, 12, cache=False)
        assert not df.empty

    def test_download_2008(self):
        df = download("SE", 2008, 6, cache=False)
        assert not df.empty

    def test_download_2010(self):
        df = download("SE", 2010, 6, cache=False)
        assert not df.empty

    def test_download_2019(self):
        df = download("SE", 2019, 6, cache=False)
        assert not df.empty


if __name__ == "__main__":
    unittest.main()



import unittest

import pandas as pd

from pysus.online_data.CNES import download


class CNESTestCase(unittest.TestCase):
    def test_fetch_estabelecimentos(self):
        df = download(group="ST", state="SP", year=2021, month=8)
        self.assertIsInstance(df, pd.DataFrame)
        # self.assertEqual(True, False)  # add assertion here

    def test_fetch_equipamentos(self):
        df = download("EQ", "RO", 2021, 9)
        self.assertIsInstance(df, pd.DataFrame)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pandas as pd

from pysus.online_data.ESUS import download


def cases_by_age_and_sex(UF, start="2020-03-01", end="2020-08-31"):
    """
    Fetches ESUS covid line list and aggregates by age and sex returning these counts between start and end dates.
    :param UF: State code
    :param start: Start date
    :param end: end date
    :return: dataframe
    """
    df = download(uf=UF)

    # Transformando as colunas em datetime type
    for cname in df:
        if cname.startswith("data"):
            df[cname] = pd.to_datetime(df[cname], errors="coerce")

    # Eliminando os valores nulos nas colunas com datas importantes
    old_size = len(df)
    df.dropna(
        subset=["dataNotificacao", "dataInicioSintomas", "dataTeste"], inplace=True
    )
    print(
        f"Removed {old_size - len(df)} rows with missing dates of symptoms, notification or testing"
    )

    # Desconsiderando os resultados negativos ou inconclusivos
    df = df.loc[~df.resultadoTeste.isin(["Negativo", "Inconclusivo ou Indeterminado"])]

    # Removendo sexo indeterminado
    df = df.loc[df.sexo.isin(["Masculino", "Feminino"])]

    # determinando a data dos primeiros sintomas como a data do index

    df["datesint"] = df["dataInicioSintomas"]
    df.set_index("datesint", inplace=True)
    df.sort_index(inplace=True, ascending=True)

    # vamos limitar a data inicial e a data final considerando apenas a primeira onda

    df = df.loc[start:end]

    ini = np.arange(0, 81, 5)
    fin = np.arange(5, 86, 5)
    fin[-1] = 120
    faixa_etaria = {f"[{i},{f})": (i, f) for i, f in zip(ini, fin)}

    labels = list(faixa_etaria.keys())
    df["faixa_etaria"] = [labels[i - 1] for i in np.digitize(df.idade, bins=ini)]

    agreg = df[["sexo", "faixa_etaria"]].groupby(["faixa_etaria", "sexo"]).size()
    agreg = agreg.reset_index()
    agreg.columns = ["faixa_etaria", "sexo", "n"]
    return agreg



import unittest

import pandas as pd
from numpy import dtype
from pysus.online_data import last_update


class TestInitFunctions(unittest.TestCase):
    def test_last_update(self):
        for db in [
            "SINAN",
            "SIM",
            "SINASC",
            "SIH",
            "SIA",
            "PNI",
            "CNES",
            "CIHA",
        ]:
            df = last_update(db)
            self.assertIsInstance(df, pd.DataFrame)
            self.assertGreater(df.size, 0)
            self.assertIn("folder", df.columns)
            self.assertIsInstance(df["date"][0], pd.Timestamp)
            self.assertEqual(df.file_size.dtype, dtype("int64"))


if __name__ == "__main__":
    unittest.main()



"""
Created on 16/08/16
by fccoelho
license: GPL V3 or Later
"""
import csv
import gzip
import os
from tempfile import NamedTemporaryFile

import pandas as pd
from dbfread import DBF
from pyreaddbc import ffi, lib
from tqdm import tqdm


def read_dbc(filename, encoding="utf-8", raw=False):
    """
    Opens a DATASUS .dbc file and return its contents as a pandas
    Dataframe.
    :param filename: .dbc filename
    :param encoding: encoding of the data
    :param raw: |
        Skip type conversion. Set it to True to avoid type conversion errors
    :return: Pandas Dataframe.
    """
    if isinstance(filename, str):
        filename = filename.encode()
    with NamedTemporaryFile(delete=False) as tf:
        dbc2dbf(filename, tf.name.encode())
        try:
            dbf = DBF(tf.name, encoding=encoding, raw=raw)
            df = pd.DataFrame(list(dbf))
        except ValueError:
            dbf = DBF(tf.name, encoding=encoding, raw=not raw)
            df = pd.DataFrame(list(dbf))
        except Exception as e:
            print(f"Failed to read DBF: {e}")
            df = pd.DataFrame()
    os.unlink(tf.name)

    return df


def dbc2dbf(infile, outfile):
    """
    Converts a DATASUS dbc file to a DBF database saving it to `outfile`.
    :param infile: .dbc file name
    :param outfile: name of the .dbf file to be created.
    """
    if isinstance(infile, str):
        infile = infile.encode()
    if isinstance(outfile, str):
        outfile = outfile.encode()
    p = ffi.new("char[]", os.path.abspath(infile))
    q = ffi.new("char[]", os.path.abspath(outfile))

    lib.dbc2dbf([p], [q])

    # print(os.path.exists(outfile))


def read_dbc_dbf(filename: str):
    if filename.endswith(("dbc", "DBC")):
        df = read_dbc(filename, encoding="iso-8859-1")
    elif filename.endswith(("DBF", "dbf")):
        dbf = DBF(filename, encoding="iso-8859-1")
        # dbf = gpd.read_file(
        # filename, encoding="iso-8859-1"
        # ).drop("geometry", axis=1)
        df = pd.DataFrame(list(dbf))
    return df


def dbf_to_csvgz(filename: str, encoding: str = "iso-8859-1"):
    """
    Streams a dbf file to gzipped CSV file. The Gzipped csv
        will be saved on the same path but with a csv.gz extension.
    :param filename: path to the dbf file
    """
    data = DBF(filename, encoding=encoding, raw=False)
    fn = os.path.splitext(filename)[0] + ".csv.gz"

    with gzip.open(fn, "wt") as gzf:
        for i, d in tqdm(
            enumerate(data),
            desc="Converting",
        ):
            if i == 0:
                csvwriter = csv.DictWriter(gzf, fieldnames=d.keys())
                csvwriter.writeheader()
                csvwriter.writerow(d)
            else:
                csvwriter.writerow(d)



"""
Download data from the national immunization program
"""
import os
import pandas as pd

from dbfread import DBF
from loguru import logger
from ftplib import FTP, error_perm

from pysus.online_data import CACHEPATH


def download(state, year, cache=True):
    """
    Download imunization records for a given State and year.
    :param state: uf two letter code
    :param year: year in 4 digits
    :param cache: If True reads from cache if available
    :return: Dataframe
    """
    # if year < 2000:
    #     raise ValueError("PNI does not contain data before 2000")
    year2 = str(year)[-2:].zfill(2)
    state = state.upper()
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")
    ftp.cwd("/dissemin/publicos/PNI/DADOS")
    logger.debug("Changing FTP work dir to: /dissemin/publicos/PNI/DADOS")
    fname = f"CPNI{state}{year2}.DBF"

    cachefile = os.path.join(CACHEPATH, "PNI_" + fname.split(".")[0] + "_.parquet")
    if os.path.exists(cachefile):
        logger.info(f"Local parquet data found at {cachefile}")
        df = pd.read_parquet(cachefile)
        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)
    except error_perm:
        try:
            ftp.retrbinary("RETR {}".format(fname.upper()), open(fname, "wb").write)
        except Exception as e:
            raise Exception("{}\nFile {} not available".format(e, fname))
    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))
    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")
    os.unlink(fname)
    logger.debug(f"{fname} removed")
    return df


def get_available_years(state):
    """
    Fetch available years (dbf names) for the `state`.
    :param state: uf code
    :return: list of strings (filenames)
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")
    ftp.cwd("/dissemin/publicos/PNI/DADOS")
    logger.debug("Changing FTP work dir to: /dissemin/publicos/PNI/DADOS")
    res = ftp.nlst(f"CPNI{state}*.DBF")
    return res


def available_docs():
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")
    ftp.cwd("/dissemin/publicos/PNI/DOCS")
    logger.debug("Changing FTP work dir to: /dissemin/publicos/PNI/DOCS")
    res = ftp.nlst(f"*")
    return res


def fetch_document(fname):
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")
    ftp.cwd("/dissemin/publicos/PNI/DOCS")
    logger.debug("Changing FTP work dir to: /dissemin/publicos/PNI/DOCS")
    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)
        print(f"Downloaded {fname}.")
    except Exception as e:
        raise Exception(f"{e}\nFile {fname} not available.")



"""
Created on 21/09/18
by fccoelho
license: GPL V3 or Later
"""
import logging
import os
import shutil
from ftplib import FTP
from pathlib import Path, PosixPath

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from dbfread import DBF
from pysus.utilities.readdbc import dbc2dbf, read_dbc

CACHEPATH = os.getenv(
    "PYSUS_CACHEPATH", os.path.join(str(Path.home()), "pysus")
)

# create pysus cache directory
if not os.path.exists(CACHEPATH):
    os.mkdir(CACHEPATH)


def cache_contents():
    """
    List the files currently cached in ~/pysus
    :return:
    """
    cached_data = os.listdir(CACHEPATH)
    return [os.path.join(CACHEPATH, f) for f in cached_data]


def _fetch_file(
    fname: str, 
    path: str,
    ftype: str, 
    return_df: bool = True, 
    data_path: str = '/tmp/pysus'
) -> pd.DataFrame:
    """
    Fetch a single file.
    :param fname: Name of the file
    :param path: ftp path where file is located
    :param ftype: 'DBC' or 'DBF'
    :return:
    Pandas Dataframe
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    ftp.cwd(path)

    Path(data_path).mkdir(exist_ok=True)

    try:
        ftp.retrbinary(f"RETR {fname}", open(f'{Path(data_path) / fname}', "wb").write)
    except Exception:
        raise Exception("File {} not available on {}".format(fname, path))
    if return_df:
        df = get_dataframe(fname, ftype, data_path)
        return df
    else:
        return pd.DataFrame()


def get_dataframe(fname: str, ftype: str, data_path: str = '/tmp/pysus') -> pd.DataFrame:
    """
    Return a dataframe read fom temporary file on disk.
    :param fname: temporary file name
    :param ftype: 'DBC' or 'DBF'
    :return:  DataFrame
    """
    fname = Path(data_path) / fname

    if ftype == "DBC":
        df = read_dbc(fname, encoding="iso-8859-1", raw=False)
    elif ftype == "DBF":
        dbf = DBF(fname, encoding="iso-8859-1", raw=False)
        df = pd.DataFrame(list(dbf))
    if os.path.exists(fname):
        os.unlink(fname)
    df.applymap(
        lambda x: x.decode("iso-8859-1") if isinstance(x, bytes) else x
    )
    return df


def chunk_dbfiles_into_parquets(fpath: str) -> str(PosixPath):

    dbfile = str(Path(fpath).absolute()).split("/")[-1]

    if Path(dbfile).suffix in [".dbc", ".DBC"]:
        outpath = f"{fpath[:-4]}.dbf"

        try:
            dbc2dbf(fpath, outpath)

        except Exception as e:
            logging.error(e)

        fpath = outpath

    parquet_dir = f"{fpath[:-4]}.parquet"
    if not Path(parquet_dir).exists():
        Path(parquet_dir).mkdir(exist_ok=True, parents=True)
        for d in stream_DBF(DBF(fpath, encoding="iso-8859-1", raw=True)):
            try:
                df = pd.DataFrame(d)
                table = pa.Table.from_pandas(
                    df.applymap(
                    lambda x: x.decode(encoding="iso-8859-1") if isinstance(x, bytes) else x
                ))
                pq.write_to_dataset(table, root_path=parquet_dir)

            except Exception as e:
                logging.error(e)

        logging.info(f"{fpath} chunked into parquets at {parquet_dir}")

    return parquet_dir


def parquets_to_dataframe(
    parquet_dir: str(PosixPath), 
    clean_after_read=False
) -> pd.DataFrame:

    parquets = Path(parquet_dir).glob("*.parquet")

    try:
        chunks_list = [
            pd.read_parquet(str(f), engine="fastparquet") for f in parquets
        ]

        return pd.concat(chunks_list, ignore_index=True)

    except Exception as e:
        logging.error(e)

    finally:
        if clean_after_read:
            shutil.rmtree(parquet_dir)
            logging.info(f"{parquet_dir} removed")


def stream_DBF(dbf, chunk_size=30000):
    """Fetches records in chunks to preserve memory"""
    data = []
    i = 0
    for records in dbf:
        data.append(records)
        i += 1
        if i == chunk_size:
            yield data
            data = []
            i = 0
    else:
        yield data


def get_CID10_table(cache=True):
    """
    Fetch the CID10 table
    :param cache:
    :return:
    """
    fname = "CID10.DBF"
    cachefile = os.path.join(
        CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet"
    )
    if os.path.exists(cachefile):
        df = pd.read_parquet(cachefile)
        return df
    df = _fetch_file(fname, "/dissemin/publicos/SIM/CID10/TABELAS", "DBF")
    if cache:
        df.to_parquet(cachefile)
    return df


DB_PATHS = {
    "SINAN": [
        "/dissemin/publicos/SINAN/DADOS/FINAIS",
        "/dissemin/publicos/SINAN/DADOS/PRELIM",
    ],
    "SIM": [
        "/dissemin/publicos/SIM/CID10/DORES",
        "/dissemin/publicos/SIM/CID9/DORES",
    ],
    "SINASC": [
        "/dissemin/publicos/SINASC/NOV/DNRES",
        "/dissemin/publicos/SINASC/ANT/DNRES",
    ],
    "SIH": [
        "/dissemin/publicos/SIHSUS/199201_200712/Dados",
        "/dissemin/publicos/SIHSUS/200801_/Dados",
    ],
    "SIA": [
        "/dissemin/publicos/SIASUS/199407_200712/Dados",
        "/dissemin/publicos/SIASUS/200801_/Dados",
    ],
    "PNI": ["/dissemin/publicos/PNI/DADOS"],
    "CNES": ["dissemin/publicos/CNES/200508_/Dados/"],
    "CIHA": ["/dissemin/publicos/CIHA/201101_/Dados"],
}


def last_update(database: str = "SINAN") -> pd.DataFrame:
    """
    Return the date of last update from the database specified.

    Parameters
    ----------
    database: Database to check
    """
    if database not in DB_PATHS:
        print(
            f"Database {database} not supported try one of these"
            "{list(DB_PATHS.keys())}"
        )
        return pd.DataFrame()

    with FTP("ftp.datasus.gov.br") as ftp:
        ftp.login()
        response = {"folder": [], "date": [], "file_size": [], "file_name": []}

        def parse(line):
            data = line.strip().split()
            response["folder"].append(pth)
            response["date"].append(
                pd.to_datetime(" ".join([data[0], data[1]]))
            )
            response["file_size"].append(
                0 if data[2] == "<DIR>" else int(data[2])
            )
            response["file_name"].append(data[3])

        for pth in DB_PATHS[database]:
            ftp.cwd(pth)
            flist = ftp.retrlines("LIST", parse)
    return pd.DataFrame(response)



u"""
Download data from CIHA and CIH (Old)
Hospital and Ambulatorial information system
http://ciha.datasus.gov.br/CIHA/index.php?area=03

Created on 12/12/18
by fccoelho
license: GPL V3 or Later
"""

import os
import pandas as pd

from dbfread import DBF
from loguru import logger
from ftplib import FTP, error_perm
from pysus.online_data import CACHEPATH
from pysus.utilities.readdbc import read_dbc


def download(state: str, year: int, month: int, cache: bool = True) -> object:
    """
    Download CIHA records for state, year and month and returns dataframe
    :param month: 1 to 12
    :param state: 2 letter state code
    :param year: 4 digit integer
    """
    state = state.upper()
    year2 = str(year)[-2:]
    month = str(month).zfill(2)
    if year < 2008:
        raise ValueError("CIHA does not contain data before 2008")
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")

    if year > 2008 and year < 2011:
        ftype = "DBC"
        ftp.cwd("/dissemin/publicos/CIH/200801_201012/Dados")
        logger.debug("Changing FTP work dir to: /dissemin/publicos/CIH/200801_201012/Dados")
        fname = "CR{}{}{}.dbc".format(state, year2, month)

    if year >= 2011:
        ftype = "DBC"
        ftp.cwd("/dissemin/publicos/CIHA/201101_/Dados")
        logger.debug("Changing FTP work dir to: /dissemin/publicos/CIHA/201101_/Dados")
        fname = "CIHA{}{}{}.dbc".format(state, str(year2).zfill(2), month)
        
    cachefile = os.path.join(CACHEPATH, "CIHA_" + fname.split(".")[0] + "_.parquet")

    if os.path.exists(cachefile):
        logger.info(f"Local parquet data found at {cachefile}")
        df = pd.read_parquet(cachefile)
        return df

    df = _fetch_file(fname, ftp, ftype)

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")
    return df


def _fetch_file(fname, ftp, ftype):
    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)
    except error_perm:
        raise Exception("File {} not available".format(fname))
    if ftype == "DBC":
        df = read_dbc(fname, encoding="iso-8859-1")
    elif ftype == "DBF":
        dbf = DBF(fname, encoding="iso-8859-1")
        df = pd.DataFrame(list(dbf))
    os.unlink(fname)
    return df



"""
Created on 17/08/16
by fccoelho
license: GPL V3 or Later
"""

import unittest
from pathlib import Path

import pandas as pd
from pysus.utilities.readdbc import read_dbc, read_dbc_dbf

PATH_ROOT = Path(__file__).resolve().parent
TEST_DATA = PATH_ROOT / "test_data"


class TestReadDbc(unittest.TestCase):
    dbf_fname = TEST_DATA / "EPR-2016-06-01-2016.dbf"
    dbc_fname = TEST_DATA / "sids.dbc"

    @unittest.skip('Issue #111')
    def test_read_dbc(self):
        df = read_dbc(str(self.dbc_fname))
        self.assertIsInstance(df, pd.DataFrame)
        self.assertGreater(df.size, 0)

    def test_read_dbf(self):
        df = read_dbc_dbf(str(self.dbf_fname))
        self.assertIsInstance(df, pd.DataFrame)
        self.assertGreater(df.size, 0)



import unittest

import pandas as pd

from pysus.online_data.PNI import *


class PNITestCase(unittest.TestCase):
    def test_get_available_years(self):
        res = get_available_years("AC")
        self.assertIsInstance(res, list)
        self.assertIn("CPNIAC00.DBF", res)

    def test_get_available_docs(self):
        res = available_docs()
        self.assertIsInstance(res, list)

    def test_fetch_doc(self):
        res = available_docs()
        fetch_document(res[0])

    def test_download(self):
        df = download("RO", 2000)
        self.assertIsInstance(df, pd.DataFrame)


if __name__ == "__main__":
    unittest.main()



"""
Helper functions to download official statistics from IBGE SIDRA
"""

import urllib3
import requests
import pandas as pd
# requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS = 'ALL:@SECLEVEL=1'

from urllib.error import HTTPError

APIBASE = "https://servicodados.ibge.gov.br/api/v3/"


def get_sidra_table(
        table_id,
        territorial_level,
        geocode="all",
        period=None,
        variables=None,
        classification=None,
        categories=None,
        format=None,
        decimals=None,
        headers=None,
):
    """
    Wrapper for the SIDRA API. More information here: http://apisidra.ibge.gov.br/home/ajuda
    :param table_id: código da tabela de onde se deseja extrair os dados. código pode ser obtido aqui: https://sidra.ibge.gov.br/acervo#/S/Q
    :param territorial_level: 1 – Brasil, 2 – Grande Região, 3 – Unidade da Federação, 6 – Município, etc
    :param geocode: geocódigo do IBGE: 3304557,3550308 – especifica os municípios do Rio de Janeiro e São Paulo.
    all – especifica todos os municípios. in n3 11,12 - especifica os municípios contidos nas Unidades da Federação Rondônia e Acre.

    :param period:  Os períodos podem ser especificados de forma avulsa, separados por vírgula (,), em faixas, separados por traço (-), ou de ambas as formas
        Um período pode ter o formato AAAA, de 4 dígitos, que representa um ano, ou o formato AAAADD, de 6 dígitos, onde AAAA representa um ano e DD seu correspondente mês (01 a 12), trimestre (01 a 04), semestre (01 a 02), etc, de acordo com a periodicidade de divulgação dos dados da tabela.

        Exemplo 1: /p/2008,2010-2012 – especifica os anos de 2008, e 2010 a 2012.

        Exemplo 2: /p/201101-201112,201204,201208 – especifica os meses de janeiro a dezembro de 2012, abril de 2012 e agosto de 2012.

        O parâmetro p pode ser seguido pela constante all para especificar todos os períodos disponíveis.

        Exemplo 3: /p/all

        O parâmetro p pode ser seguido pela constante first e um número de períodos, indicando os primeiros períodos da lista de períodos disponíveis (períodos mais antigos).
        O número de períodos pode ser omitido quando se tratar de apenas um aperíodo.

        Exemplo 4: /p/first 12

        Exemplo 5: /p/first

        O parâmetro p pode ser seguido pela constante last e um número de períodos, indicando os últimos períodos da série (períodos mais recentes).
        O número de períodos pode ser omitido quando se tratar de apenas um período.

        Exemplo 6: /p/last 12

        Exemplo 7: /p/last (valor default, quando não especificado o parâmetro p)
    :param variables:As variáveis são especificadas através de seus códigos, separados por vírgula (,).
        A lista de variáveis pode incluir também as variáveis de percentual geradas automaticamente pelo Sidra (são variáveis cujos códigos são superiores a 1.000.000).

        Exemplo 1: /v/63,69 – especifica o percentual no mês e o percentual acumulado no ano do IPCA.

        O parâmetro v pode ser seguido pela constante all para especificar todas as variáveis da tabela, inclusive as variáveis de percentual geradas automaticamente pelo Sidra.

        Exemplo 2: /v/all

        O parâmetro v pode ser seguido pela constante allxp para especificar todas as variáveis da tabela, exceto as variáveis de percentual geradas automaticamente pelo Sidra.

        Exemplo 3: /v/allxp (valor default, quando não especificado o parâmetro v)
    :param classification: informa o código de uma das classificações da tabela.
        Como exemplos, temos 1 – Situação do domicílio, 2 – Sexo, 81 – Produto da lavoura temporária, etc.
    :param categories:  As categorias são especificadas através de seus códigos, de forma individual ou para compor uma soma, separadas por vírgula (,).
        As categorias que compõem a soma devem ser separadas por espaço.

        Exemplo 1: /c81/2692,2702,2694 2695 – especifica os produtos da lavoura temporária arroz, feijão e (batata doce + batata inglesa)
    :param format:
    :param decimals:
    :param headers: `y` para receber o header (valor default, caso o parâmetro h não seja especificado). `n` para não receber o header.
    :return:
    """
    base_url = "https://apisidra.ibge.gov.br/values"
    query = f"/t/{table_id}/n{territorial_level}/{geocode}"
    if period is not None:
        query += f"/p/{period}"
    if variables is not None:
        query += f"/v/{variables}"
    if classification is not None:
        query += f"/c{classification}"
    if categories is not None:
        query += f"/{categories}"
    if format is not None:
        query += f"/f/{format}"
    if decimals is not None:
        query += f"/d/{decimals}"
    if headers is not None:
        query += f"/h/{headers}"

    url = base_url + query
    print(f"Requesting data from {url}")
    try:
        with (
            get_legacy_session() as s,
            s.get(url) as response
        ):
            df = pd.DataFrame(response.json())
    except HTTPError as exc:
        response = requests.get(url)
        print(f"Consulta falhou: {response.text}")
        return None
    return df


def list_agregados(**kwargs):
    """
    Lista de agregados agrupados por pesquisa.
    veja https://servicodados.ibge.gov.br/api/docs/agregados?versao=3#api-Agregados-agregadosGet
    para maiores detalhes
    :param kwargs: parâmetros válidos: período, assunto, classificacao, periodicidade,nivel.
    :return: Dataframe
    """
    url = APIBASE + "agregados?"
    url += "&".join([f"{k}={v}" for k, v in kwargs.items()])
    print(f"Fetching Data groupings from {url}")
    try:
        with (
            get_legacy_session() as s,
            s.get(url) as response
        ):
            table = pd.DataFrame(response.json())
    except requests.exceptions.SSLError as e:
        print(f"Failed fetching aggregates: {e}")
        return pd.DataFrame()
    return table


def localidades_por_agregado(agregado: int, nivel: str):
    """
    Obtém as localidades associadas ao agregado de acordo com um ou mais níveis geográficos.
    :param agregado: codigo numérico do agregado
    :param nivel: Identificador do nível geográfico ao qual pertence as localidades. Pode conter um ou mais níveis
    delimitados pelo caracter | (pipe). p.ex. N7|N6
    :return:
    """
    url = APIBASE + f"agregados/{agregado}/localidades/{nivel}"
    try:
        with (
            get_legacy_session() as s,
            s.get(url) as response
        ):
            table = pd.DataFrame(response.json())
    except Exception as e:
        print(f"Could not download from {url}\n{e}")
        return None
    return table


def metadados(agregado: int):
    """
    Obtém os metadados associados ao agregado

    :param agregado: Identificador do agregado
    """
    url = APIBASE + f"agregados/{agregado}/metadados"
    try:
        with (
            get_legacy_session() as s,
            s.get(url) as response
        ):
            data = response.json()
    except Exception as e:
        print(f"Could not download from {url}\n{e}")
        return None
    return data


def lista_periodos(agregado: int):
    """
    Obtém os períodos associados ao agregado
    :param agregado:
    :return: pd.DataFrame com os períodos de atualização
    """
    url = APIBASE + f"agregados/{agregado}/periodos"
    try:
        with (
            get_legacy_session() as s,
            s.get(url) as response
        ):
            table = pd.DataFrame(response.json())
    except:
        return None
    return table


class FetchData:
    """
    Obtém o conjunto de variáveis a partir do identificador do agregado, períodos pesquisados e identificador das variáveis
    :param agregado: identifocador do agregados
    :param periodos: Período do qual se deseja obter os resultados. Consulte os identificadores dos períodos na Base de
        identificadores. Informe valores negativos para obter os últimos resultados. Pode conter um ou mais períodos
        delimitados pelo caracter | (pipe)
    :param variavel: Um ou mais identificadores de variável separados pelo caracter | (pipe). Caso omitido, assume o
        valor allxp, que retorna quaisquer variáveis relacionada ao agregado. Para saber mais sobre as variáveis de cada
        agregado, acesse seus respectivos metadados
    :kwargs: parametros adicionais:
        - **localidades**: Uma ou mais localidades delimitadas pelo caracter | (pipe). No caso do Brasil, o identificador é
            BR. Para qualquer outra localidade que NÃO seja Brasil, essa deve seguir o padrão N<NIVEL_GEOGRAFICO>[<LOCALIDADE>],
            em que <LOCALIDADE> pode ser uma ou mais localidades separadas por vírgula. É possível ainda generalizar o
            resultado, informando a classe da localidade, conforme os exemplos a seguir

                https://servicodados.ibge.gov.br/api/v3/agregados/1705/variaveis?localidades=N7

            Obtém os resultados referentes às variáveis do agregado 1705 cujas localidades sejam regiões metropolitanas (N7)

                https://servicodados.ibge.gov.br/api/v3/agregados/1705/variaveis?localidades=N7[3501,3301]

            Obtém os resultados referentes às variáveis do agregado 1705 cujas localidades sejam as regiões metropolitanas
            (N7) de São Paulo e Rio de Janeiro (3501,3301). Observe que 3501 e 3301 são, respectivamente, os identificadores
            das regiões metropolitanas de São Paulo e Rio de Janeiro. Não podem ser confundidos, portanto, com os
            identificadores dos municípios de São Paulo/SP e Rio de Janeiro/RJ, que são 3550308 e 3304557, respectivamente

        - **classificacao**: Além de estar relacionado à uma dada localidade e um determinado período, os resultados das
            variáveis podem estar relacionados à outros conjuntos de dados, que na nomenclatura do SIDRA recebe o nome de
            classificação. Como exemplo, considere o agregado Produção, venda, valor da produção e área colhida da lavoura
            temporária nos estabelecimentos agropecuários. Além da localidade e do período, os resultados produzidos por
            esse agregado referem-se aos produtos produzidos, condição do produtor, grupos de atividades econômica, grupos
            de área, grupos de área colhida e pronafiano, que são as classificações do agregado - Para conhecer as
            classificações de cada agregado, acesse seus respectivos metadados. Aos componentes da classificação, dar-se o
            nome de categoria. Na prática, você fará uso das classificações para restringir a consulta, conforme os exemplos a seguir

            https://servicodados.ibge.gov.br/api/v3/agregados/1712/variaveis?classificacao=226[4844]&localidades=BR

            Obtém os resultados referentes às variáveis do agregado 1712 cujo produto produzido (226) seja abacaxi (4844)
            no Brasil (BR)

            https://servicodados.ibge.gov.br/api/v3/agregados/1712/variaveis?classificacao=226[4844]|218[4780]&localidades=BR

            Obtém os resultados referentes às variáveis do agregado 1712 cujo produto produzido (226) seja abacaxi (4844) e
            cuja condição do produtor (218) seja proprietário (4780) no Brasil (BR)
        - **view**: Modo de visualização. Caso deseje que a resposta seja renderizada usando notação OLAP, configure
            esse parâmetro com o valor OLAP - https://servicodados.ibge.gov.br/api/v3/agregados/1705/variaveis?view=OLAP&localidades=BR.
            A outra opção é configurar esse parâmetro com o valor flat. No modo flat, o primeiro elemento do Array são
            metadados, de forma que os resultados vêm a partir do segundo elemento
    """

    def __init__(self, agregado: int, periodos: str, variavel: str = "allxp", **kwargs):
        self.url = (
                APIBASE + f"agregados/{agregado}/periodos/{periodos}/variaveis/{variavel}?"
        )
        self.url += "&".join([f"{k}={v}" for k, v in kwargs.items()])
        self.JSON = None
        self._fetch_JSON()

    def _fetch_JSON(self):
        try:
            print(f"Fetching {self.url}")
            with (
                get_legacy_session() as s,
                s.get(self.url) as response
            ):
                self.JSON = response.json()
        except Exception as e:
            print(f"Couldn't download data:\n{e}")

    def to_dataframe(self):
        return pd.DataFrame(self.JSON)



"""
HTTPSConnectionPool(host='servicodados.ibge.gov.br', port=443): 
    Max retries exceeded with url: 
    /api/v3/agregados/{...} 
    Caused by SSLError(
        SSLError(1, '[SSL: UNSAFE_LEGACY_RENEGOTIATION_DISABLED] 
        unsafe legacy renegotiation disabled (_ssl.c:1129)'

SOLUTION: https://github.com/scrapy/scrapy/issues/5491#issuecomment-1241862323
"""

import ssl # Builtin


class CustomHttpAdapter (requests.adapters.HTTPAdapter):
    # "Transport adapter" that allows us to use custom ssl_context.

    def __init__(self, ssl_context=None, **kwargs):
        self.ssl_context = ssl_context
        super().__init__(**kwargs)

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = urllib3.poolmanager.PoolManager(
            num_pools=connections, maxsize=maxsize,
            block=block, ssl_context=self.ssl_context)


def get_legacy_session():
    ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
    ctx.options |= 0x4  # OP_LEGACY_SERVER_CONNECT
    session = requests.session()
    session.mount('https://', CustomHttpAdapter(ctx))
    return session



import json
import string
from difflib import get_close_matches
from pathlib import Path
from typing import Dict

import pandas as pd
import unidecode

# from loguru import logger

APP_DIR = Path(__file__).resolve(strict=True).parent.parent
CID10 = {"dengue": "A90", "chikungunya": "A92.0", "zika": "A928"}

with open(APP_DIR / "dataset/geocode_by_cities.json", "r") as f:
    geocode_by_cities = json.load(f)


def normalize(s):
    for p in string.punctuation:
        s = s.replace(p, '')

    return unidecode.unidecode(s.lower().strip())


def search_string(substr: str) -> Dict[str, int]:
    """
    Fetch geocode of the city name matching to the substring.

    Parameters
    ----------
        substr: Part of city name
    Returns
    -------
        dict: Dictionary with key and values
            with city name and IBGE codes of all municipalities in Brazil
    """
    normalized_list = [normalize(f) for f in list(geocode_by_cities.keys())]
    
    matching_cities = [
        get_close_matches(
            i, normalized_list, n=55
        )
        for i in normalize(substr).split(".")
    ]

    return {
        key: geocode_by_cities[key]
        for key in geocode_by_cities
        if normalize(key) in list(*matching_cities)
    }


def download(
    disease: str,
    eyw_start: int,
    eyw_end: int,
    city_name: str,
    format="csv",
) -> pd.DataFrame:
    """
    Download InfoDengue API data by municipality and disease
        in the epidemiological week.

    Parameters
    ----------
        disease: Names of the diseases available in the InfoDengue System:
            dengue|chikungunya|zika
        eyw_start: Epidemiological week start
        eyw_end: Epidemiological week end
        city_name: Name of the municipalities of Brazil
        format="csv": Default data visualization format for the endpoint
    Returns
    -------
        pd: Pandas dataframe
    """

    geocode = geocode_by_cities.get(city_name)

    if disease not in CID10.keys():
        raise Exception(
            f"The diseases available are: {[k for k in CID10.keys()]}"
        )
    elif len(str(eyw_start)) != 6 or len(str(eyw_end)) != 6:
        raise Exception(
            "The epidemiological week must contain 6 digits, "
            "started in the year 2010 until 2022. Example: 202248"
        )
    elif geocode is None:
        list_of_cities = search_string(city_name)
        print(f"You must choose one of these city names: {list_of_cities}")
    else:
        s_yw = str(eyw_start)
        e_yw = str(eyw_end)
        ew_start, ey_start = s_yw[-2:], s_yw[:4]
        ew_end, ey_end = e_yw[-2:], e_yw[:4]
        url = "https://info.dengue.mat.br/api/alertcity"
        params = (
            "&disease="
            + f"{disease}"
            + "&geocode="
            + f"{geocode}"
            + "&format="
            + f"{format}"
            + "&ew_start="
            + f"{ew_start}"
            + "&ew_end="
            + f"{ew_end}"
            + "&ey_start="
            + f"{ey_start}"
            + "&ey_end="
            + f"{ey_end}"
        )

        url_resp = "?".join([url, params])

        return pd.read_csv(url_resp, index_col="SE").T



__author__ = "fccoelho"

import unittest

import pandas as pd

from pysus.online_data.CIHA import download

unittest.skip("too slow to run om travis")


class SIHTestCase(unittest.TestCase):
    def test_download_CIH(self):
        df = download("mg", 2009, 7)

        self.assertGreater(len(df), 0)
        self.assertIn("DIAG_PRINC", df.columns)
        self.assertIsInstance(df, pd.DataFrame)

    def test_download_CIHA(self):
        df = download("MG", 2013, 10)
        self.assertGreater(len(df), 0)
        self.assertIn("DIAG_PRINC", df.columns)
        self.assertIsInstance(df, pd.DataFrame)


if __name__ == "__main__":
    unittest.main()



import os
from functools import lru_cache

import geocoder
import numpy as np
import pandas as pd
import requests
from dbfread import DBF


def read_sinan_dbf(fname, encoding) -> pd.DataFrame:
    """
    Read SINAN dbf file returning a Pandas Dataframe with
    :param fname: dbf file name
    :param encoding: Encoding of the dbf
    :return: pandas dataframe
    """
    db = DBF(fname, encoding=encoding)
    df = pd.DataFrame(list(db))

    def convert_week(x):
        try:
            w = int(x) % 100
        except ValueError:
            w = np.nan
        return w

    for cname in df.columns:
        df[cname].replace("", np.nan, inplace=True)
        if cname.startswith(("NU", "ID")):
            try:
                df[cname] = pd.to_numeric(df[cname])
            except ValueError as e:
                print(f"Column {cname} could not be converted to numeric: {e}")
                # certain IDs can be alphanumerical
                pass
        elif cname.startswith("SEM"):
            df[cname] = df[cname].map(convert_week)

    return df


@lru_cache(maxsize=None)
def get_geocodes(geoc):
    """
    Return city name and state two letter code from geocode
    :param geoc:
    :return:
    """
    url = (
        "http://cidades.ibge.gov.br/services/jSonpMuns.php?"
        "busca=330&featureClass=P&style=full&maxRows=5&name_startsWith={}".format(geoc)
    )
    resp = requests.get(url)
    for d in resp.json()["municipios"]:
        if int(geoc) == int(d["c"]):
            return [d["n"].encode("latin-1").decode("utf-8"), d["s"]]

    else:
        raise KeyError("could not find geocode {} in ".format(geoc))


def _address_generator(df, default=""):
    for row in df.iterrows():
        line = dict(row[1])
        try:
            line["cidade"] = ",".join(get_geocodes(line["ID_MN_RESI"]))
        except KeyError:
            print("Could not find geocode {} using default")
            line["cidade"] = default
        yield line[
            "NU_NOTIFIC"
        ], "{NM_LOGRADO}, {NU_NUMERO}, {NM_BAIRRO}, {cidade}, Brasil".format(**line)


def geocode(sinan_df, outfile, default_city):
    """
    Geocode cases based on addresses included.
    :param default_city: default city to use in case of bad Geocode found in file. It can be "city, state"
    :param sinan_df: Dataframe generated from sinan DBF
    :param outfile: File on Which
    """
    addrs = _address_generator(sinan_df, default_city)
    if os.path.exists(outfile):
        mode = "a"
        coords = pd.read_csv(outfile)
        geocoded = coords.NU_NOTIFIC.tolist()
    else:
        mode = "w"
        geocoded = []
    with open(outfile, mode) as of:
        if mode == "w":
            of.write("NU_NOTIFIC,latitude,longitude\n")
        for nu, ad in addrs:
            # ad = ad.encode('latin-1').decode('utf-8')
            if nu in geocoded:
                continue
            location = geocoder.google(ad)
            if location is None:
                raise NameError("Google could not find {}".format(ad))
            if location.latlng == []:
                print(
                    "Search for {} returned {} as coordinates, trying reduced address:".format(
                        ad, location.latlng
                    )
                )
                ad = ",".join(ad.split(",")[2:])
                print(ad)
                location = geocoder.google(ad)
            try:
                of.write(
                    "{},{},{}\n".format(nu, location.latlng[0], location.latlng[1])
                )
                print("Successfully geolocated {}".format(ad))
            except IndexError:
                print(
                    "Search for {} returned {} as coordinates, skipping".format(
                        ad, location.latlng
                    )
                )
                of.write("{},nan,nan\n".format(nu))



u"""
Download Mortality records from SIM Datasus
Created on 12/12/18
by fccoelho
license: GPL V3 or Later
"""

import os
import pandas as pd

from dbfread import DBF
from loguru import logger
from ftplib import FTP, error_perm

from pysus.online_data import CACHEPATH
from pysus.utilities.readdbc import read_dbc


def download(state, year, cache=True, folder=None):
    """
    Downloads data directly from Datasus ftp server
    :param state: two-letter state identifier: MG == Minas Gerais
    :param year: 4 digit integer
    :return: pandas dataframe
    """
    year2 = str(year)[-2:].zfill(2)
    state = state.upper()
    ftp_dir = ""
    fname = ""

    if year < 1979:
        raise ValueError("SIM does not contain data before 1979")

    elif year >= 1996:
        ftp_dir = "/dissemin/publicos/SIM/CID10/DORES"
        fname = "DO{}{}.DBC".format(state, year)

    else:
        ftp_dir = "/dissemin/publicos/SIM/CID9/DORES"
        fname = fname = "DOR{}{}.DBC".format(state, year2)

    cache_fail = False
    cachefile = os.path.join(CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet")

    if folder:
        fname = "{}/{}".format(folder, fname)

    elif cache:
        if os.path.exists(cachefile):
            logger.info(f"Local parquet file found at {cachefile}")
            df = pd.read_parquet(cachefile)

            return df

        else:
            cache_fail = True

    # Se tiver folder não tenta cache
    if not folder and (cache_fail or not cache):
        ftp = FTP("ftp.datasus.gov.br")
        ftp.login()
        logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")
        ftp.cwd(ftp_dir)
        logger.debug(f"Changing FTP work dir to: {ftp_dir}")

        try:
            ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

        except error_perm:
            try:
                ftp.retrbinary("RETR {}".format(fname.upper()), open(fname, "wb").write)

            except:
                raise Exception("File {} not available".format(fname))

    df = read_dbc(fname, encoding="iso-8859-1")

    df.to_parquet(cachefile)
    logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")
    return df


def get_CID10_chapters_table(cache=True):
    """
    Fetch the CID10 chapters table
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")
    ftp.cwd("/dissemin/publicos/SIM/CID10/TABELAS")
    logger.debug("Changing FTP work dir to: /dissemin/publicos/SIM/CID10/TABELAS")

    fname = "CIDCAP10.DBF"
    cachefile = os.path.join(CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet")
    
    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df
        
    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except error_perm:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def get_CID10_table(cache=True):
    """
    Fetch the CID10 table
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")
    ftp.cwd("/dissemin/publicos/SIM/CID10/TABELAS")
    logger.debug("Changing FTP work dir to: /dissemin/publicos/SIM/CID10/TABELAS")

    fname = "CID10.DBF"
    cachefile = os.path.join(CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet")

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except error_perm:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def get_CID9_table(cache=True):
    """
    Fetch the CID9 table
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")
    ftp.cwd("/dissemin/publicos/SIM/CID9/TABELAS")
    logger.debug("Changing FTP work dir to: /dissemin/publicos/SIM/CID9/TABELAS")

    fname = "CID9.DBF"
    cachefile = os.path.join(CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet")

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except error_perm:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def get_municipios(cache=True):
    """
    Get municipality metadata
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")
    ftp.cwd("/dissemin/publicos/SIM/CID10/TABELAS")
    logger.debug("Changing FTP work dir to: /dissemin/publicos/SIM/CID10/TABELAS")

    fname = "CADMUN.DBF"
    cachefile = os.path.join(CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet")

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df
        
    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def get_ocupations(cache=True):
    """
    Fetch ocupations table
    :param cache: If set to True, stores data as parquets.
    :return: Pandas DataFrame
    """
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")
    ftp.cwd("/dissemin/publicos/SIM/CID10/TABELAS")
    logger.debug("Changing FTP work dir to: /dissemin/publicos/SIM/CID10/TABELAS")
    fname = "TABOCUP.DBF"
    cachefile = os.path.join(CACHEPATH, "SIM_" + fname.split(".")[0] + "_.parquet")

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except:
        raise Exception("Could not download {}".format(fname))

    dbf = DBF(fname, encoding="iso-8859-1")
    df = pd.DataFrame(list(dbf))

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df



"""
Downloads SIA data from Datasus FTP server
Created on 21/09/18
by fccoelho
Modified on 22/11/22
by bcbernardo
license: GPL V3 or Later
"""

import os
import pandas as pd

from ftplib import FTP
from datetime import date
from loguru import logger
from pprint import pprint
from typing import Dict, List, Optional, Tuple, Union

from pysus.online_data import CACHEPATH
from pysus.utilities.readdbc import read_dbc_dbf, dbc2dbf

group_dict: Dict[str, Tuple[str, int, int]] = {
    "PA": ("Produção Ambulatorial", 7, 1994),
    "BI": ("Boletim de Produção Ambulatorial individualizado", 1, 2008),
    "AD": ("APAC de Laudos Diversos", 1, 2008),
    "AM": ("APAC de Medicamentos", 1, 2008),
    "AN": ("APAC de Nefrologia", 1, 2008),
    "AQ": ("APAC de Quimioterapia", 1, 2008),
    "AR": ("APAC de Radioterapia", 1, 2008),
    "AB": ("APAC de Cirurgia Bariátrica", 1, 2008),
    "ACF": ("APAC de Confecção de Fístula", 1, 2008),
    "ATD": ("APAC de Tratamento Dialítico", 1, 2008),
    "AMP": ("APAC de Acompanhamento Multiprofissional", 1, 2008),
    "SAD": ("RAAS de Atenção Domiciliar", 1, 2008),
    "PS": ("RAAS Psicossocial", 1, 2008),
}

def show_datatypes():
    pprint(group_dict)

def download(
        state: str,
        year: int,
        month: int,
        cache: bool = True,
        group: Union[str, List[str]] = ["PA", "BI"],
) -> Union[Optional[pd.DataFrame], Tuple[Optional[pd.DataFrame], ...]]:
    """
    Download SIASUS records for state year and month and returns dataframe
    :param month: 1 to 12
    :param state: 2 letter state code
    :param year: 4 digit integer
    :param cache: whether to cache files locally. default is True
    :param groups: 2-3 letter document code or a list of 2-3 letter codes,
        defaults to ['PA', 'BI']. Codes should be one of the following:
        PA - Produção Ambulatorial
        BI - Boletim de Produção Ambulatorial individualizado
        AD - APAC de Laudos Diversos
        AM - APAC de Medicamentos
        AN - APAC de Nefrologia
        AQ - APAC de Quimioterapia
        AR - APAC de Radioterapia
        AB - APAC de Cirurgia Bariátrica
        ACF - APAC de Confecção de Fístula
        ATD - APAC de Tratamento Dialítico
        AMP - APAC de Acompanhamento Multiprofissional
        SAD - RAAS de Atenção Domiciliar
        PS - RAAS Psicossocial
    :return: A tuple of dataframes with the documents in the order given
        by the , when they are found
    """
    state = state.upper()
    year2 = str(year)[-2:]
    month = str(month).zfill(2)
    if isinstance(group, str):
        group = [group]
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")
    ftype = "DBC"
    if year >= 1994 and year < 2008:
        ftp.cwd("/dissemin/publicos/SIASUS/199407_200712/Dados")
        logger.debug("Changing FTP work dir to: /dissemin/publicos/SIASUS/199407_200712/Dados")

    elif year >= 2008:
        ftp.cwd("/dissemin/publicos/SIASUS/200801_/Dados")
        logger.debug("Changing FTP work dir to: /dissemin/publicos/SIASUS/200801_/Dados")

    else:
        raise ValueError("SIA does not contain data before 1994")

    dfs: List[Optional[pd.DataFrame]] = []
    for gname in group:
        gname = gname.upper()
        if gname not in group_dict:
            raise ValueError(f"SIA does not contain files named {gname}")

        # Check available
        input_date = date(int(year), int(month), 1)
        available_date = date(group_dict[gname][2], group_dict[gname][1], 1)
        if input_date < available_date:
            dfs.append(None)
            # NOTE: raise Warning instead of ValueError for
            # backwards-compatibility with older behavior of returning
            # (PA, None) for calls after 1994 and before Jan, 2008
            logger.warning(
                f"SIA does not contain data for {gname} "
                f"before {available_date:%d/%m/%Y}"
            )
            continue

        fname = f"{gname}{state}{year2.zfill(2)}{month}.dbc"

        # Check in Cache
        cachefile = os.path.join(CACHEPATH, "SIA_" + fname.split(".")[0] + "_.parquet")
        if os.path.exists(cachefile):
            logger.info(f"Local parquet file found at {cachefile}")
            df = pd.read_parquet(cachefile)
        else:
            try:
                df = _fetch_file(fname, ftp, ftype)
                if cache and df:  # saves to cache if df is not None
                    df.to_parquet(cachefile)
                    logger.info(f"Data stored as parquet at {cachefile}")
            except Exception as e:
                df = None
                print(e)

        dfs.append(df)

    return tuple(dfs)


def _fetch_file(fname, ftp, ftype):
    """
    Does the FTP fetching.
    :param fname: file name
    :param ftp: ftp connection object
    :param ftype: file type: DBF|DBC
    :return: pandas dataframe
    """

    multiples = False
    fnames = check_file_split(fname, ftp)

    multiples = len(fnames) > 1

    if multiples:
        download_multiples(fnames, ftp)
        print(f"This download is split into the following files: {fnames}\n"
              f"They have been downloaded in {CACHEPATH}.\n"
              f"To load them, use the pysus.utilities.read_dbc_dbf function.")
        return
    df = read_dbc_dbf(fname)

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df


def download_multiples(fnames, ftp):
    for fn in fnames:
        fnfull = os.path.join(CACHEPATH, fn)
        print(f"Downloading {fn}...")
        fobj = open(fnfull, "wb")
        try:
            ftp.retrbinary(f"RETR {fn}", fobj.write)
            dbc2dbf(fnfull, fnfull.replace('.dbc', '.dbf'))
            os.unlink(fnfull)
            logger.debug(f"{fnfull} removed")
        except Exception as exc:
            raise Exception(f"Retrieval of file {fn} failed with the following error:\n {exc}")


def check_file_split(fname: str, ftp: FTP) -> list:
    """
    Check for split filenames. Sometimes when files are too large, they are split into multiple files ending in a, b, c, ...
    :param fname: filename
    :param ftp: ftp conection
    :return: list
    """
    files = []
    flist = ftp.nlst()
    if fname not in flist:
        for l in ['a', 'b', 'c', 'd']:
            nm, ext = fname.split('.')
            if f'{nm}{l}.{ext}' in flist:
                files.append(f'{nm}{l}.{ext}')

    return files



import os
import pandas as pd

from dbfread import DBF
from loguru import logger
from datetime import datetime
from ftplib import FTP, error_perm

from pysus.online_data import CACHEPATH
from pysus.utilities.readdbc import read_dbc

group_dict = {
    "LT": ["Leitos - A partir de Out/2005", 10, 2005],
    "ST": ["Estabelecimentos - A partir de Ago/2005", 8, 2005],
    "DC": ["Dados Complementares - A partir de Ago/2005", 8, 2005],
    "EQ": ["Equipamentos - A partir de Ago/2005", 8, 2005],
    "SR": ["Serviço Especializado - A partir de Ago/2005", 8, 2005],
    "HB": ["Habilitação - A partir de Mar/2007", 3, 2007],
    "PF": ["Profissional - A partir de Ago/2005", 8, 2005],
    "EP": ["Equipes - A partir de Abr/2007", 5, 2007],
    "IN": ["Incentivos - A partir de Nov/2007", 11, 2007],
    "RC": ["Regra Contratual - A partir de Mar/2007", 3, 2007],
    "EE": ["Estabelecimento de Ensino - A partir de Mar/2007", 3, 2007],
    "EF": ["Estabelecimento Filantrópico - A partir de Mar/2007", 3, 2007],
    "GM": ["Gestão e Metas - A partir de Jun/2007", 6, 2007],
}


def download(
        group: str, state: str, year: int, month: int, cache: bool = True
) -> pd.DataFrame:
    """
    Download CNES records for group, state, year and month and returns dataframe
    :param group:
        LT – Leitos - A partir de Out/2005
        ST – Estabelecimentos - A partir de Ago/2005
        DC - Dados Complementares - A partir de Ago/2005
        EQ – Equipamentos - A partir de Ago/2005
        SR - Serviço Especializado - A partir de Ago/2005
        HB – Habilitação - A partir de Mar/2007
        PF – Profissional - A partir de Ago/2005
        EP – Equipes - A partir de Abr/2007
        IN – Incentivos - A partir de Nov/2007
        RC - Regra Contratual - A partir de Mar/2007
        EE - Estabelecimento de Ensino - A partir de Mar/2007
        EF - Estabelecimento Filantrópico - A partir de Mar/2007
        GM - Gestão e Metas - A partir de Jun/2007
    :param month: 1 to 12
    :param state: 2 letter state code
    :param year: 4 digit integer
    """
    state = state.upper()
    assert len(str(year)) == 4
    year2 = str(year)[-2:]
    month = str(month).zfill(2)
    input_date = datetime(int(year), int(month), 1)
    avaiable_date = datetime(group_dict[group][2], group_dict[group][1], 1)

    if input_date < avaiable_date:
        raise ValueError(f"CNES does not contain data for {input_date}")

    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")

    if input_date >= avaiable_date:
        ftype = "DBC"
        ftp.cwd("dissemin/publicos/CNES/200508_/Dados/{}/".format(group))
        logger.debug("Changing FTP work dir to: dissemin/publicos/CNES/200508_/Dados/{}/".format(group))
        fname = "{}{}{}{}.dbc".format(group, state, str(year2).zfill(2), month)

    cachefile = os.path.join(CACHEPATH, "CNES_" + fname.split(".")[0] + "_.parquet")

    if os.path.exists(cachefile):
        logger.info(f"Local parquet data found at {cachefile}")
        df = pd.read_parquet(cachefile)
        return df

    df = _fetch_file(fname, ftp, ftype)

    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    return df


def _fetch_file(fname: str, ftp: FTP, ftype: str) -> pd.DataFrame:
    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)
    except error_perm:
        raise Exception("File {} not available".format(fname))
    if ftype == "DBC":
        df = read_dbc(fname, encoding="iso-8859-1")
    elif ftype == "DBF":
        dbf = DBF(fname, encoding="iso-8859-1")
        df = pd.DataFrame(list(dbf))
    os.unlink(fname)
    logger.debug(f"{fname} removed.")
    return df



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pysus',
 'pysus.online_data',
 'pysus.preprocessing',
 'pysus.tests',
 'pysus.tests.test_data',
 'pysus.utilities']

package_data = \
{'': ['*'], 'pysus': ['Notebooks/*', 'dataset/*']}

install_requires = \
['Unidecode>=1.3.6,<2.0.0',
 'cffi==1.15.1',
 'dbfread==2.0.7',
 'elasticsearch>=8.3.3,<9.0.0',
 'fastparquet>=0.8.1,<0.9.0',
 'geocoder>=1.38.1,<2.0.0',
 'jupyterlab>=3.4.5,<4.0.0',
 'loguru>=0.6.0,<0.7.0',
 'numpy==1.23.2',
 'pandas==1.4.3',
 'pyarrow>=9.0.0,<10.0.0',
 'pycparser==2.21',
 'pyreaddbc==1.0.0',
 'python-dateutil==2.8.2',
 'pytz==2022.2.1',
 'six==1.16.0',
 'tqdm==4.64.0',
 'wget>=3.2,<4.0']

setup_kwargs = {
    'name': 'pysus',
    'version': '0.7.0',
    'description': "Tools for dealing with Brazil's Public health data",
    'long_description': None,
    'author': 'Flavio Codeco Coelho',
    'author_email': 'fccoelho@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# -*- coding:utf-8 -*-
u"""
Created on 19/07/16
by fccoelho
license: GPL V3 or Later
"""

import unittest

import numpy as np
import pandas as pd
from numpy.testing import *

from pysus.online_data.SIM import download, get_CID10_chapters_table
from pysus.preprocessing import decoders
from pysus.preprocessing.SIM import (
    group_and_count,
    redistribute_cid_chapter,
    redistribute_missing,
)


def get_CID10_code(index, code):
    try:
        code = index[code]
    except KeyError:
        code = -1
    return code


class TestDecoder(unittest.TestCase):
    def test_decodifica_idade_retorna_em_anos(self):
        res = decoders.decodifica_idade_SINAN(4010, unidade="Y")
        self.assertEqual(res, 10)
        res = decoders.decodifica_idade_SINAN(3120, unidade="Y")
        self.assertEqual(res, 10)
        res = decoders.decodifica_idade_SINAN(2365, unidade="Y")
        self.assertAlmostEqual(res, 1, places=4)
        res = decoders.decodifica_idade_SINAN(1480, unidade="Y")
        self.assertAlmostEqual(res, 0.0547, places=3)

    def test_decodifica_lista_idades_retorna_em_anos(self):
        res = decoders.decodifica_idade_SINAN([4010] * 3, unidade="Y")
        assert_array_equal(res, np.array([10] * 3))
        res = decoders.decodifica_idade_SINAN([3120] * 4, unidade="Y")
        assert_array_equal(res, np.array([10] * 4))
        res = decoders.decodifica_idade_SINAN([2365] * 2, unidade="Y")
        assert_array_equal(res, np.array([1, 1]))
        res = decoders.decodifica_idade_SINAN([1480] * 5, unidade="Y")
        assert_array_almost_equal(res, np.array([0.0547] * 5), decimal=3)

    def test_decodifica_idade_retorna_em_anos_SIM(self):
        res = decoders.decodifica_idade_SIM(["501"], unidade="Y")
        assert_array_equal(res, np.array([101]))
        res = decoders.decodifica_idade_SIM(["401"] * 2, unidade="Y")
        assert_array_equal(res, np.array([1] * 2))
        res = decoders.decodifica_idade_SIM(["311"] * 3, unidade="Y")
        assert_array_almost_equal(res, np.array([0.904109589] * 3), decimal=3)
        res = decoders.decodifica_idade_SIM(["224"] * 4, unidade="Y")
        assert_array_almost_equal(res, np.array([0.065753425] * 4), decimal=3)
        res = decoders.decodifica_idade_SIM(["130"] * 5, unidade="Y")
        assert_array_almost_equal(res, np.array([0.00274] * 5), decimal=3)
        res = decoders.decodifica_idade_SIM(["010"] * 6, unidade="m")
        assert_array_almost_equal(res, np.array([10.0] * 6))

    def test_verifica_geocodigo(self):
        self.assertTrue(decoders.is_valid_geocode(3304557))

    def test_translate_variables(self):
        df = download("sp", 2010)
        df = decoders.translate_variables_SIM(df)
        sex_array = df["SEXO"].unique().tolist()
        assert_array_equal(sex_array, ["Masculino", "Feminino", "nan"])
        raca_array = df["RACACOR"].unique().tolist()
        assert_array_equal(
            raca_array, ["Branca", "Preta", "Amarela", "nan", "Parda", "Indígena"]
        )

    def test_get_cid_chapter(self):
        code_index = decoders.get_CID10_code_index(get_CID10_chapters_table())
        test_causes = pd.DataFrame(
            {
                "causas": [
                    "A00",
                    "B99",
                    "D48",
                    "D49",
                    "D50",
                    "H00",
                    "H59",
                    "H60",
                    "V00",
                    "W00",
                    "X00",
                    "U00",
                    "U04",
                ]
            }
        )
        results = test_causes["causas"].map(lambda x: get_CID10_code(code_index, x))
        assert_array_equal(results, [1, 1, 2, -1, 3, 7, 7, 8, -1, 20, 20, -1, 22])

    def test_group_and_count(self):
        df = download("se", 2010)
        df = decoders.translate_variables_SIM(df)
        variables = ["CODMUNRES", "SEXO", "IDADE_ANOS"]
        counts = group_and_count(df, variables)
        sample = (
            counts[counts["COUNTS"] != 0]["COUNTS"].sample(20, random_state=0).tolist()
        )
        self.assertGreater(sum(sample), 0)

    def test_redistribute(self):
        df = download("sp", 2010)
        df = decoders.translate_variables_SIM(
            df, age_classes=True, classify_cid10_chapters=True
        )
        variables = ["CODMUNRES", "SEXO", "IDADE_ANOS", "CID10_CHAPTER"]
        df = df[variables]
        counts = group_and_count(df, variables)
        sum_original = counts["COUNTS"].sum()
        counts = redistribute_missing(counts, variables)
        sum_redistributed = counts["COUNTS"].sum()

        assert_almost_equal(sum_original, sum_redistributed, 10)

        sample = (
            counts[counts["COUNTS"] != 0]["COUNTS"].sample(20, random_state=0).tolist()
        )
        assert_array_almost_equal(
            sample,
            [
                1.0,
                1.0000216033775462,
                4.0,
                1.0057015548341106,
                2.000363538647316,
                3.0005453079709743,
                1.0,
                2.0093748859678917,
                1.0,
                1.0006631753413024,
                1.0,
                1.0155903470702614,
                1.0006446228186379,
                1.0007163086475952,
                4.0016700388384105,
                1.0003146522751405,
                5.202681974105347,
                1.0057015548341106,
                1.0006806444217275,
                1.0000656718488452,
            ],
            decimal=5,
        )

        counts = redistribute_cid_chapter(counts, ["CODMUNRES", "SEXO", "IDADE_ANOS"])
        sum_redistributed = counts["COUNTS"].sum()

        assert_almost_equal(sum_original, sum_redistributed, 10)

        sample = (
            counts[counts["COUNTS"] != 0]["COUNTS"].sample(20, random_state=0).tolist()
        )
        assert_array_almost_equal(
            sample,
            [
                1.089135695829918,
                1.1471212205224637,
                97.66379391566016,
                1.0006806444217275,
                1.0526404291598292,
                1.0002258989870523,
                1.0006438895125183,
                1.0022096833374972,
                1.004692969527825,
                1.0098947488581271,
                1.3848786564718214,
                1.0358818448712763,
                1.0477163671352119,
                1.1041264089747516,
                1.0002258989870523,
                4.00889998546595,
                1.0435326872735615,
                4.000315617188721,
                1.0007163086475952,
                2.0118196033377975,
            ],
            decimal=5,
        )



"""
Downloads SIH data from Datasus FTP server
Created on 21/09/18
by fccoelho
license: GPL V3 or Later
"""

import os
import pandas as pd

from ftplib import FTP
from dbfread import DBF
from loguru import logger

from pysus.online_data import CACHEPATH
from pysus.utilities.readdbc import read_dbc


def download(state: str, year: int, month: int, cache: bool = True) -> object:
    """
    Download SIH records for state year and month and returns dataframe
    :param month: 1 to 12
    :param state: 2 letter state code
    :param year: 4 digit integer
    :param cache: Whether to cache or not. defaults to True.
    :return:
    """
    state = state.upper()
    year2 = int(str(year)[-2:])
    year2 = str(year2).zfill(2)
    month = str(month).zfill(2)

    if year < 1992:
        raise ValueError("SIH does not contain data before 1994")

    if year < 2008:
        ftype = "DBC"
        path = "/dissemin/publicos/SIHSUS/199201_200712/Dados"
        fname = f"RD{state}{year2}{month}.dbc"

    if year >= 2008:
        ftype = "DBC"
        path = f"/dissemin/publicos/SIHSUS/200801_/Dados"
        fname = f"RD{state}{year2}{month}.dbc"

    cachefile = os.path.join(CACHEPATH, "SIH_" + fname.split(".")[0] + "_.parquet")

    if os.path.exists(cachefile):
        logger.info(f"Local parquet file found at {cachefile}")
        df = pd.read_parquet(cachefile)

        return df

    df = _fetch_file(fname, path, ftype)
    
    if cache:
        df.to_parquet(cachefile)
        logger.info(f"Data stored as parquet at {cachefile}")

    return df


def _fetch_file(fname, path, ftype):
    ftp = FTP("ftp.datasus.gov.br")
    ftp.login()
    logger.debug(f"Stablishing connection with ftp.datasus.gov.br.\n{ftp.welcome}")
    ftp.cwd(path)
    logger.debug(f"Changing FTP work dir to: {path}")

    try:
        ftp.retrbinary("RETR {}".format(fname), open(fname, "wb").write)

    except:
        raise Exception("File {} not available".format(fname))

    if ftype == "DBC":
        df = read_dbc(fname, encoding="iso-8859-1")

    elif ftype == "DBF":
        dbf = DBF(fname, encoding="iso-8859-1")
        df = pd.DataFrame(list(dbf))

    os.unlink(fname)
    logger.debug(f"{fname} removed")

    return df



__author__ = "fccoelho"

import unittest

from pysus.online_data.sinasc import download, get_available_years


class TestDownload(unittest.TestCase):
    def test_download_new(self):
        df = download("SE", 2015)
        self.assertIn("IDADEMAE", df.columns)
        self.assertGreater(len(df), 0)

    def test_download_old(self):
        df = download("AL", 1994)
        self.assertIn("IDADE_MAE", df.columns)
        self.assertGreater(len(df), 0)

    def test_get_available_years(self):
        yrs = get_available_years("AC")
        self.assertIn("DNAC1996.DBC", yrs)
        self.assertIn("DNRAC94.DBC", yrs)


if __name__ == "__main__":
    unittest.main()
